/** Globals for generator
 *
 * @author Ondrej Cekan
 * @date 20/02/2014
 * @update 05/07/2015
 *
     Copyright (C) 2015 Ondrej Cekan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <vector>
#include <map>
#include <climits>

/**< Default values */
unsigned CPBALANCE = 0;     // balance for line count
unsigned MAXROWS = 25000;   // maximum of lines
unsigned MINROWS = 25000;   // minimum of lines
unsigned IMAX = MAXROWS;    // number of instructions to generate
const unsigned UNSIGNED_MAX = ULONG_MAX-100000; // max unsigned value
unsigned NSLOTS = 1;        // number of slots
string DELIMITER = "\n";    // delimiter of instructions



using namespace std;

/**< Global variables */
/**< For processor description */
vector<Tsubs> subs;
vector<Tsubs>::iterator subsIt;

map<string,Tsyn> syn;
map<string,Tsyn>::iterator synIt;

vector<pair<string, unsigned> > storeval;
vector<pair<string, unsigned> >::iterator storevalIt;


/**< For processor constraints */
vector<Tvar> var;
vector<Tvar>::iterator varIt;

/**< Global constraints */
vector<Tconsit> gcons;
vector<Tconsit>::iterator gconsIt;
vector<Tconsit>::reverse_iterator gconsItr;

vector<pair<string, unsigned> > rate;   // vector of assigned substitutes
vector<pair<string, unsigned> >::iterator rateIt;

vector<pair<string, string> > repl;     // forced replacing with substitutes
vector<pair<string, string> >::iterator replIt;
vector<pair<string, string> >::reverse_iterator replItr;




/** Test string for containing specified chars
 *
 * @param word - tested word
 * @return boolean
 *
 */
bool is_keyword(string word)
{
    for (string::iterator it = word.begin(); it != word.end(); ++it)
    {
        if (!isalnum(*it) && *it != '_') return false;
    }

    return true;
}


/** Test if string is integer number
 *
 * @param word - string
 * @return boolean
 *
 */
bool is_numeric(string word)
{
    bool e = false;
    for (string::iterator it = word.begin(); it != word.end(); ++it)
    {
        if (*it == 'e') e = true;
        else if ((!isdigit(*it) && *it != '+' && *it != '-') || e) return false;
    }

    return true;
}


/** Test if constraint is valid
 *
 * @param word - string
 * @return boolean
 *
 */
bool is_constraint(string word)
{
    if ((word.compare("pb") == 0) ||
        (word.compare("contain") == 0) ||
        (word.compare("unique") == 0) ||
        (word.compare("different") == 0) ||
        (word.compare("outall") == 0) ||
        (word.compare("nouse") == 0) ||
        (word.compare("afterinsert") == 0) ||
        (word.compare("beforeinsert") == 0) ||
        (word.compare("nlines") == 0) ||
        (word.compare("minval") == 0) ||
        (word.compare("maxval") == 0) ||
        (word.compare("mul") == 0) ||
        (word.compare("div") == 0) ||
        (word.compare("mod") == 0) ||
        (word.compare("mdiv") == 0) ||
        (word.compare("start") == 0) ||
        (word.compare("end") == 0) ||
        (word.compare("useonly") == 0) ||
        (word.compare("nslots") == 0) ||
        (word.compare("slot") == 0) ||
        (word.compare("nocare") == 0) ||
        (word.compare("annulcomposed") == 0) ||
        (word.compare("nolinecount") == 0) ||
        (word.compare("linedelimiter") == 0) ||
        (word.compare("cposbalance") == 0) ||
        (word.compare("storeval") == 0) ||
        (word.compare("output") == 0)) return true;
    return false;
}


/** Replace all phrases in string
 *
 * @param str - string
 * @param from - word that is replaced
 * @param to - word being replaced
 * @return replaced string
 *
 */
string replace_str(string str, string from, string to)
{
    unsigned spos = str.find(from);
    while(spos != string::npos)
    {
        str.replace(spos, from.length(), to);
        spos = str.find(from);
    }

    return str;
}

/** Replace phrase in string
 *
 * @param str - string
 * @param from - word that is replaced
 * @param to - word being replaced
 * @return replaced string
 *
 */
string replace_one(string str, string from, string to)
{
    unsigned spos = str.find(from);
    if(spos != string::npos)
    {
        str.replace(spos, from.length(), to);
    }

    return str;
}


/** Replace all same phrases in string
 *
 * @param str - string
 * @param from - word that is replaced
 * @param to - word being replaced
 * @return replaced string
 *
 */
string replace_str_all(string str, string from, string to)
{
    while (1)
    {
        string res = replace_str(str, from, to);
        if (res.compare(str) == 0) return res;
        else str = res;
    }
}


/** Transfer int to string
 *
 * @param number - input number
 * @return string
 *
 */
string intToString(int number)
{
   stringstream ss;
   ss << number;
   return ss.str();
}


/** Counts number of lines in string
 *
 * @param str - string
 * @return number of lines
 *
 */
int numOfLines(string str)
{
    int num = 0;
    size_t found = -1;

    do
    {
        found = str.find(DELIMITER,found+1);
        if (found != string::npos)
        {
            num++;
        }
    } while (found != string::npos);

    if (num == 0)
    {
        found = -1;

        do
        {
            found = str.find("\n",found+1);
            if (found != string::npos)
            {
                num++;
            }
        } while (found != string::npos);
    }

    return num;
}


/** Split string into vector
 *
 * @param s - input string
 * @param delim - delimiter
 * @param elems - output elements of the string
 * @return elements
 *
 */
vector<string> &split(const string &s, char delim, vector<string> &elems) {
    stringstream ss(s);
    string item;
    while (getline(ss, item, delim)) {
        elems.push_back(item);
    }
    return elems;
}


/** Split string into vector
 *
 * @param s - input string
 * @param delim - delimiter
 * @return elements
 *
 */
vector<string> split(const string &s, char delim) {
    vector<string> elems;
    split(s, delim, elems);
    return elems;
}

